@@ -3,7 +3,7 @@ from typing import TYPE_CHECKING, Iterable, Optional
 from attrs import define
 
 if TYPE_CHECKING:
-    from .index import BaseDataIndex
+    from .index import BaseDataIndex, DataIndexKey
 
 from .index import DataIndexEntry
 
@@ -20,6 +20,20 @@ class Change:
     old: Optional[DataIndexEntry]
     new: Optional[DataIndexEntry]
 
+    @property
+    def key(self) -> "DataIndexKey":
+        if self.typ == RENAME:
+            raise ValueError
+
+        if self.typ == ADD:
+            entry = self.new
+        else:
+            entry = self.old
+
+        assert entry
+        assert entry.key
+        return entry.key
+
     def __bool__(self):
         return self.typ != UNCHANGED
 
@@ -72,9 +86,20 @@ def _detect_renames(changes: Iterable[Change]):
         else:
             yield change
 
+    def _get_key(change):
+        return change.key
+
+    added[:] = sorted(added, key=_get_key)
+    deleted[:] = sorted(deleted, key=_get_key)
+
     for change in added:
         new_entry = change.new
         assert new_entry
+
+        if not new_entry.hash_info:
+            yield change
+            continue
+
         index, old_entry = None, None
         for idx, ch in enumerate(deleted):
             assert ch.old
